"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.MongoDbFactory = exports.MongoDb = exports.MongoCollection = void 0;
const console_1 = require("console");
const mongodb_1 = require("mongodb");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const MaxFetchSize = 2000;
class MongoCollection {
    constructor(collection) {
        this.collection = collection;
    }
    aggregate(pipeline, options) {
        return this.collection.aggregate(pipeline, options);
    }
    // eslint-disable-next-line @typescript-eslint/ban-types,@typescript-eslint/promise-function-async
    find(query, sort, limit = MaxFetchSize, skip) {
        let queryCursor = this.collection
            .find(query)
            .sort(sort)
            .limit(limit);
        if (skip) {
            queryCursor = queryCursor.skip(skip);
        }
        return queryCursor.toArray();
    }
    // eslint-disable-next-line @typescript-eslint/ban-types,@typescript-eslint/promise-function-async
    findOne(query) {
        return this.collection.findOne(query);
    }
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    findAll() {
        return this.collection.find({}).toArray();
    }
    // eslint-disable-next-line @typescript-eslint/ban-types
    async update(filter, set, addToSet) {
        return this.updateCore(filter, set, addToSet, false);
    }
    // eslint-disable-next-line @typescript-eslint/ban-types
    async updateMany(filter, set, addToSet) {
        return this.updateManyCore(filter, set, addToSet, false);
    }
    // eslint-disable-next-line @typescript-eslint/ban-types
    async upsert(filter, set, addToSet) {
        return this.updateCore(filter, set, addToSet, true);
    }
    async distinct(key, query) {
        return this.collection.distinct(key, query);
    }
    async deleteOne(filter) {
        return this.collection.deleteOne(filter);
    }
    async deleteMany(filter) {
        return this.collection.deleteMany(filter);
    }
    async insertOne(value) {
        const result = await this.collection.insertOne(value);
        return result.insertedId;
    }
    async insertMany(values, ordered) {
        await this.collection.insertMany(values, { ordered: false });
    }
    async createIndex(index, unique) {
        try {
            const indexName = await this.collection.createIndex(index, { unique });
            server_services_telemetry_1.Lumberjack.info(`Created index ${indexName}`);
        }
        catch (error) {
            server_services_telemetry_1.Lumberjack.error(`Index creation failed`, error);
        }
    }
    async createTTLIndex(index, expireAfterSeconds) {
        await this.collection.createIndex(index, { expireAfterSeconds });
    }
    async findOrCreate(query, value) {
        const result = await this.collection.findOneAndUpdate(query, {
            $setOnInsert: value,
        }, {
            returnOriginal: true,
            upsert: true,
        });
        return result.value
            ? { value: result.value, existing: true }
            : { value, existing: false };
    }
    async updateCore(filter, set, addToSet, upsert) {
        const update = {};
        if (set) {
            update.$set = set;
        }
        if (addToSet) {
            update.$addToSet = addToSet;
        }
        const options = { upsert };
        await this.collection.updateOne(filter, update, options);
    }
    async updateManyCore(filter, set, addToSet, upsert) {
        const update = {};
        if (set) {
            update.$set = set;
        }
        if (addToSet) {
            update.$addToSet = addToSet;
        }
        const options = { upsert };
        await this.collection.updateMany(filter, update, options);
    }
}
exports.MongoCollection = MongoCollection;
class MongoDb {
    constructor(client) {
        this.client = client;
    }
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    close() {
        return this.client.close();
    }
    on(event, listener) {
        this.client.on(event, listener);
    }
    collection(name) {
        const collection = this.client.db("admin").collection(name);
        return new MongoCollection(collection);
    }
    async dropCollection(name) {
        return this.client.db("admin").dropCollection(name);
    }
}
exports.MongoDb = MongoDb;
class MongoDbFactory {
    constructor(config) {
        const { operationsDbEndpoint, bufferMaxEntries, globalDbEnabled, globalDbEndpoint, connectionPoolMinSize, connectionPoolMaxSize, } = config;
        if (globalDbEnabled) {
            this.globalDbEndpoint = globalDbEndpoint;
        }
        (0, console_1.assert)(!!operationsDbEndpoint, `No endpoint provided`);
        this.operationsDbEndpoint = operationsDbEndpoint;
        this.bufferMaxEntries = bufferMaxEntries;
        this.connectionPoolMinSize = connectionPoolMinSize;
        this.connectionPoolMaxSize = connectionPoolMaxSize;
    }
    async connect(global = false) {
        var _a;
        (0, console_1.assert)(!global || !!this.globalDbEndpoint, `No global endpoint provided
                 when trying to connect to global db.`);
        // Need to cast to any before MongoClientOptions due to missing properties in d.ts
        const options = {
            autoReconnect: true,
            bufferMaxEntries: (_a = this.bufferMaxEntries) !== null && _a !== void 0 ? _a : 50,
            keepAlive: true,
            keepAliveInitialDelay: 180000,
            reconnectInterval: 1000,
            reconnectTries: 100,
            socketTimeoutMS: 120000,
            useNewUrlParser: true,
        };
        if (this.connectionPoolMinSize) {
            options.minSize = this.connectionPoolMinSize;
        }
        if (this.connectionPoolMaxSize) {
            options.poolSize = this.connectionPoolMaxSize;
        }
        const connection = await mongodb_1.MongoClient.connect(global ?
            this.globalDbEndpoint :
            this.operationsDbEndpoint, options);
        return new MongoDb(connection);
    }
}
exports.MongoDbFactory = MongoDbFactory;
//# sourceMappingURL=mongodb.js.map