"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.TenantManager = exports.Tenant = void 0;
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_utils_1 = require("@fluidframework/server-services-utils");
const common_utils_1 = require("@fluidframework/common-utils");
class Tenant {
    constructor(config, manager) {
        this.config = config;
        this.manager = manager;
    }
    get id() {
        return this.config.id;
    }
    get gitManager() {
        return this.manager;
    }
    get storage() {
        return this.config.storage;
    }
    get orderer() {
        return this.config.orderer;
    }
}
exports.Tenant = Tenant;
/**
 * Manages a collection of tenants
 */
class TenantManager {
    constructor(endpoint, internalHistorianUrl) {
        this.endpoint = endpoint;
        this.internalHistorianUrl = internalHistorianUrl;
    }
    async createTenant(tenantId) {
        const restWrapper = new server_services_client_1.BasicRestWrapper();
        const result = await restWrapper.post(`${this.endpoint}/api/tenants/${encodeURIComponent(tenantId || "")}`, undefined);
        return result;
    }
    async getTenant(tenantId, documentId, includeDisabledTenant = false) {
        const restWrapper = new server_services_client_1.BasicRestWrapper();
        const [details, key] = await Promise.all([
            restWrapper.get(`${this.endpoint}/api/tenants/${tenantId}`, { includeDisabledTenant }),
            this.getKey(tenantId, includeDisabledTenant)
        ]);
        const defaultQueryString = {
            token: (0, common_utils_1.fromUtf8ToBase64)(`${tenantId}`),
        };
        const getDefaultHeaders = () => {
            const credentials = {
                password: (0, server_services_utils_1.generateToken)(tenantId, documentId, key, null),
                user: tenantId,
            };
            return ({
                Authorization: (0, server_services_client_1.getAuthorizationTokenFromCredentials)(credentials),
            });
        };
        const defaultHeaders = getDefaultHeaders();
        const baseUrl = `${this.internalHistorianUrl}/repos/${encodeURIComponent(tenantId)}`;
        const tenantRestWrapper = new server_services_client_1.BasicRestWrapper(baseUrl, defaultQueryString, undefined, undefined, defaultHeaders, undefined, undefined, getDefaultHeaders, server_services_utils_1.getCorrelationId);
        const historian = new server_services_client_1.Historian(`${this.internalHistorianUrl}/repos/${encodeURIComponent(tenantId)}`, true, false, tenantRestWrapper);
        const gitManager = new server_services_client_1.GitManager(historian);
        const tenant = new Tenant(details, gitManager);
        return tenant;
    }
    async verifyToken(tenantId, token) {
        const restWrapper = new server_services_client_1.BasicRestWrapper();
        await restWrapper.post(`${this.endpoint}/api/tenants/${encodeURIComponent(tenantId)}/validate`, { token });
    }
    async getKey(tenantId, includeDisabledTenant = false) {
        const restWrapper = new server_services_client_1.BasicRestWrapper();
        const result = await restWrapper.get(`${this.endpoint}/api/tenants/${encodeURIComponent(tenantId)}/keys`, { includeDisabledTenant });
        return result.key1;
    }
}
exports.TenantManager = TenantManager;
//# sourceMappingURL=tenant.js.map