"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Throttler = void 0;
const server_services_core_1 = require("@fluidframework/server-services-core");
const lru_cache_1 = __importDefault(require("lru-cache"));
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * A lenient implementation of IThrottlerHelper that prioritizes low latency over strict throttling.
 * This should be used for implementing throttling in places where latency matters more than accuracy,
 * such as service endpoints or socket connections.
 */
class Throttler {
    constructor(throttlerHelper, minThrottleIntervalInMs = 1000000, logger, maxCacheSize = 1000, maxCacheAge = 1000 * 60) {
        this.throttlerHelper = throttlerHelper;
        this.minThrottleIntervalInMs = minThrottleIntervalInMs;
        this.logger = logger;
        const cacheOptions = {
            max: maxCacheSize,
            maxAge: maxCacheAge,
        };
        this.lastThrottleUpdateAtMap = new lru_cache_1.default(cacheOptions);
        this.countDeltaMap = new lru_cache_1.default(cacheOptions);
        this.throttlerResponseCache = new lru_cache_1.default(cacheOptions);
    }
    /**
     * Increments operation count and calculates throttle status of given operation id.
     * Uses most recently calculated throttle status to determine current throttling, while updating in the background.
     * @throws {@link ThrottlingError} if throttled
     */
    incrementCount(id, weight = 1, usageStorageId, usageData) {
        var _a;
        this.updateCountDelta(id, weight);
        void this.updateAndCacheThrottleStatus(id, usageStorageId, usageData);
        // check cached throttle status, but allow operation through if status is not yet cached
        const cachedThrottlerResponse = this.throttlerResponseCache.get(id);
        // eslint-disable-next-line @typescript-eslint/prefer-optional-chain
        if (cachedThrottlerResponse && cachedThrottlerResponse.throttleStatus) {
            const retryAfterInSeconds = Math.ceil(cachedThrottlerResponse.retryAfterInMs / 1000);
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Throttled: ${id}`, {
                messageMetaData: {
                    key: id,
                    reason: cachedThrottlerResponse.throttleReason,
                    retryAfterInSeconds,
                    eventName: "throttling",
                },
            });
            server_services_telemetry_1.Lumberjack.info(`Throttled: ${id}`, {
                [server_services_telemetry_1.CommonProperties.telemetryGroupName]: "throttling",
                [server_services_telemetry_1.ThrottlingTelemetryProperties.key]: id,
                [server_services_telemetry_1.ThrottlingTelemetryProperties.reason]: cachedThrottlerResponse.throttleReason,
                [server_services_telemetry_1.ThrottlingTelemetryProperties.retryAfterInSeconds]: retryAfterInSeconds,
            });
            // eslint-disable-next-line @typescript-eslint/no-throw-literal
            throw new server_services_core_1.ThrottlingError(cachedThrottlerResponse.throttleReason, retryAfterInSeconds);
        }
    }
    /**
     * Decrements operation count of given operation id.
     */
    decrementCount(id, weight = 1) {
        this.updateCountDelta(id, -weight);
    }
    updateCountDelta(id, value) {
        const currentValue = this.countDeltaMap.get(id) || 0;
        this.countDeltaMap.set(id, currentValue + value);
    }
    async updateAndCacheThrottleStatus(id, usageStorageId, usageData) {
        const now = Date.now();
        if (this.lastThrottleUpdateAtMap.get(id) === undefined) {
            this.lastThrottleUpdateAtMap.set(id, now);
        }
        const lastThrottleUpdateTime = this.lastThrottleUpdateAtMap.get(id);
        if (now - lastThrottleUpdateTime > this.minThrottleIntervalInMs) {
            const countDelta = this.countDeltaMap.get(id);
            this.lastThrottleUpdateAtMap.set(id, now);
            this.countDeltaMap.set(id, 0);
            const messageMetaData = {
                key: id,
                weight: countDelta,
                eventName: "throttling",
            };
            const lumberjackProperties = {
                [server_services_telemetry_1.CommonProperties.telemetryGroupName]: "throttling",
                [server_services_telemetry_1.ThrottlingTelemetryProperties.key]: id,
                [server_services_telemetry_1.ThrottlingTelemetryProperties.weight]: countDelta,
            };
            // poplulate usageData with relevant data.
            if (usageData) {
                usageData.value = countDelta;
                usageData.startTime = lastThrottleUpdateTime;
                usageData.endTime = now;
            }
            await this.throttlerHelper.updateCount(id, countDelta, usageStorageId, usageData)
                .then((throttlerResponse) => {
                var _a;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Incremented throttle count for ${id} by ${countDelta}`, { messageMetaData });
                server_services_telemetry_1.Lumberjack.info(`Incremented throttle count for ${id} by ${countDelta}`, lumberjackProperties);
                this.throttlerResponseCache.set(id, throttlerResponse);
            })
                .catch((err) => {
                var _a;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.error(`Failed to update throttling count for ${id}: ${err}`, { messageMetaData });
                server_services_telemetry_1.Lumberjack.error(`Failed to update throttling count for ${id}`, lumberjackProperties, err);
            });
        }
    }
}
exports.Throttler = Throttler;
//# sourceMappingURL=throttler.js.map